__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2019-12-05'

import logging
import os
import json
import keras
import pickle

from src.run_modules.run_environment import RunEnvironment
from src.data_handling.data_distributor import Distributor
from src.plotting.training_monitoring import PlotModelHistory, PlotModelLearningRate
from src.model_modules.keras_extensions import LearningRateDecay, ModelCheckpointAdvanced


class Training(RunEnvironment):

    def __init__(self):
        super().__init__()
        self.model: keras.Model = self.data_store.get("model", "general.model")
        self.train_set = None
        self.val_set = None
        self.test_set = None
        self.batch_size = self.data_store.get("batch_size", "general.model")
        self.epochs = self.data_store.get("epochs", "general.model")
        self.checkpoint: ModelCheckpointAdvanced = self.data_store.get("checkpoint", "general.model")
        # self.callbacks = self.data_store.get("callbacks", "general.model")
        self.lr_sc = self.data_store.get("lr_decay", "general.model")
        self.hist = self.data_store.get("hist", "general.model")
        self.experiment_name = self.data_store.get("experiment_name", "general")
        self._run()

    def _run(self) -> None:
        """
        Perform training
        1) set_generators():
            set generators for training, validation and testing and distribute according to batch size
        2) make_predict_function():
            create predict function before distribution on multiple nodes (detailed information in method description)
        3) train():
            train model and save callbacks
        4) save_model():
            save best model from training as final model
        """
        self.set_generators()
        self.make_predict_function()
        self.train()
        self.save_model()

    def make_predict_function(self) -> None:
        """
        Creates the predict function. Must be called before distributing. This is necessary, because tf will compile
        the predict function just in the moment it is used the first time. This can cause problems, if the model is
        distributed on different workers. To prevent this, the function is pre-compiled. See discussion @
        https://stackoverflow.com/questions/40850089/is-keras-thread-safe/43393252#43393252
        """
        self.model._make_predict_function()

    def _set_gen(self, mode: str) -> None:
        """
        Set and distribute the generators for given mode regarding batch size
        :param mode: name of set, should be from ["train", "val", "test"]
        """
        gen = self.data_store.get("generator", f"general.{mode}")
        setattr(self, f"{mode}_set", Distributor(gen, self.model, self.batch_size))

    def set_generators(self) -> None:
        """
        Set all generators for training, validation, and testing subsets. The called sub-method will automatically
        distribute the data according to the batch size. The subsets can be accessed as class variables train_set,
        val_set, and test_set .
        """
        for mode in ["train", "val", "test"]:
            self._set_gen(mode)

    def train(self) -> None:
        """
        Perform training using keras fit_generator(). Callbacks are stored locally in the experiment directory. Best
        model from training is saved for class variable model.
        """
        logging.info(f"Train with {len(self.train_set)} mini batches.")
        if not os.path.exists(self.checkpoint.filepath):
            history = self.model.fit_generator(generator=self.train_set.distribute_on_batches(),
                                               steps_per_epoch=len(self.train_set),
                                               epochs=self.epochs,
                                               verbose=2,
                                               validation_data=self.val_set.distribute_on_batches(),
                                               validation_steps=len(self.val_set),
                                               callbacks=[self.lr_sc, self.hist, self.checkpoint])
        else:
            lr_filepath = self.checkpoint.callbacks[0]["path"]
            hist_filepath = self.checkpoint.callbacks[1]["path"]
            self.lr_sc = pickle.load(open(lr_filepath, "rb"))
            self.hist = pickle.load(open(hist_filepath, "rb"))
            self.model = keras.models.load_model(self.checkpoint.filepath)
            initial_epoch = max(self.hist.epoch) + 1
            callbacks = [{"callback": self.lr_sc, "path": lr_filepath},
                         {"callback": self.hist, "path": hist_filepath}]
            self.checkpoint.update_callbacks(callbacks)
            self.checkpoint.update_best(self.hist)
            _ = self.model.fit_generator(generator=self.train_set.distribute_on_batches(),
                                         steps_per_epoch=len(self.train_set),
                                         epochs=self.epochs,
                                         verbose=2,
                                         validation_data=self.val_set.distribute_on_batches(),
                                         validation_steps=len(self.val_set),
                                         callbacks=[self.lr_sc, self.hist, self.checkpoint],
                                         initial_epoch=initial_epoch)
            history = self.hist
        self.save_callbacks_as_json(history)
        self.load_best_model(self.checkpoint.filepath)
        self.create_monitoring_plots(history, self.lr_sc)

    def save_model(self) -> None:
        """
        save model in local experiment directory. Model is named as <experiment_name>_my_model.h5 .
        """
        path = self.data_store.get("experiment_path", "general")
        name = f"{self.data_store.get('experiment_name', 'general')}_my_model.h5"
        model_name = os.path.join(path, name)
        logging.debug(f"save best model to {model_name}")
        self.model.save(model_name)
        self.data_store.set("best_model", self.model, "general")

    def load_best_model(self, name: str) -> None:
        """
        Load model weights for model with name. Skip if no weights are available.
        :param name: name of the model to load weights for
        """
        logging.debug(f"load best model: {name}")
        try:
            self.model.load_weights(name)
            logging.info('reload weights...')
        except OSError:
            logging.info('no weights to reload...')

    def save_callbacks_as_json(self, history: keras.callbacks.History) -> None:
        """
        Save callbacks (history, learning rate) of training.
        * history.history -> history.json
        * lr_sc.lr -> history_lr.json
        :param history: history object of training
        """
        logging.debug("saving callbacks")
        path = self.data_store.get("experiment_path", "general")
        with open(os.path.join(path, "history.json"), "w") as f:
            json.dump(history.history, f)
        with open(os.path.join(path, "history_lr.json"), "w") as f:
            json.dump(self.lr_sc.lr, f)

    def create_monitoring_plots(self, history: keras.callbacks.History, lr_sc: LearningRateDecay) -> None:
        """
        Creates the history and learning rate plot in dependence of the number of epochs. The plots are saved in the
        experiment's plot_path. History plot is named '<exp_name>_history_loss_val_loss.pdf', the learning rate with
        '<exp_name>_history_learning_rate.pdf'.
        :param history: keras history object with losses to plot (must include 'loss' and 'val_loss')
        :param lr_sc:  learning rate decay object with 'lr' attribute
        """
        path = self.data_store.get("plot_path", "general")
        name = self.data_store.get("experiment_name", "general")

        # plot history of loss and mse (if available)
        filename = os.path.join(path, f"{name}_history_loss.pdf")
        PlotModelHistory(filename=filename, history=history)
        multiple_branches_used = len(history.model.output_names) > 1  # means that there are multiple output branches
        if multiple_branches_used:
            filename = os.path.join(path, f"{name}_history_main_loss.pdf")
            PlotModelHistory(filename=filename, history=history, main_branch=True)
        if "mean_squared_error" in history.model.metrics_names:
            filename = os.path.join(path, f"{name}_history_main_mse.pdf")
            PlotModelHistory(filename=filename, history=history, plot_metric="mse", main_branch=multiple_branches_used)

        # plot learning rate
        PlotModelLearningRate(filename=os.path.join(path, f"{name}_history_learning_rate.pdf"), lr_sc=lr_sc)
